/*	Agent.h
 *
 *	Robert Rose June 17, 2005
 *
 *	Copyright (c) 2005 Robert Rose
 *
 *	Permission is granted to do whatever you want with this code as long
 *	as the author is given due credit.
 *	
 */

#pragma once

#include <vector>

#include "NogredexDemoFrameListener.h"

using namespace std;


typedef enum {
	AgentTypePlayer = 0,
	AgentTypeEnemy,
	AgentTypeObstacle,
} AgentType;

typedef enum {
	AgentSearchStrategyNaiveGrid = 0,
	AgentSearchStrategyNaiveGrid4,
	AgentSearchStrategyRays,
	AgentSearchStrategyRayCombo,
} AgentSearchStrategy;

class AgentSearchPoint {
public:
	float x;
	float z;

	bool operator==(const AgentSearchPoint &other) {
		const float thresh = 0.5;
		return (abs(this->x - other.x) < thresh && abs(this->z - other.z) < thresh);
	}
};

class AgentPathSegment {
public:
	int i;			// direction number, prevents us from backing up
	float f;		// A* "f" - estimated cost of this solution at this node
	float g;		// A* "g" - cost to get to this node
	Vector3 path;	// direction of movement for this node

	AgentPathSegment(Vector3 &path)
	{
		this->i = 0;
		this->g = 0.0;
		this->f = 0.0;
		this->path = path;
	}
	
	bool operator<(AgentPathSegment &other)
	{
		if(f < other.f)
			return true;
		else
			return false;
	}
};

class CAgent
{
protected:
	CAgent();
public:
	CAgent(NogredexDemoFrameListener *nogre_env, AgentType type = AgentTypePlayer);
	virtual ~CAgent(void);

	Vector3 getPosition() const							{ return toOgre(_obj->actor->getGlobalPosition()); }
	void setPosition(const Vector3 &p)					{ _obj->actor->setGlobalPosition(toNx(p)); }

	void spin(const Vector3 &dir, float force = 4000.0);
	void push(const Vector3 &dir, float force = 4000.0);

	void think(float delta, const AgentSearchStrategy strat, const vector<CAgent *> &agents, const CAgent &player, const vector<CAgent *> &obstacles, vector< pair <Vector3, Vector3> > &visForces, bool applyForce);

	virtual AxisAlignedBox getBox() const				{ return AxisAlignedBox(); }
	virtual Vector3 getDistance(CAgent *other) const	{ return getPosition() - other->getPosition(); }

protected:

	Vector3 attract(const CAgent *agent, float threshold, float force);

	bool pathfind_naive(const Vector3 &pos, const CAgent &player, const vector<CAgent *> &obstacles, vector< pair <Vector3, Vector3> > &visForces, Vector3 &force, float g = 0.0, int prev = -1);
	bool pathfind_naive4(const Vector3 &pos, const CAgent &player, const vector<CAgent *> &obstacles, vector< pair <Vector3, Vector3> > &visForces, Vector3 &force, vector<AgentSearchPoint> &prev, float g = 0.0);
	bool pathfind_ray(const Vector3 &pos, const CAgent &player, const vector<CAgent *> &obstacles, vector< pair <Vector3, Vector3> > &visForces, Vector3 &force, float g = 0.0, int depth = 0);
	bool pathfind_raycombo(const Vector3 &pos, const CAgent &player, const vector<CAgent *> &obstacles, vector< pair <Vector3, Vector3> > &visForces, Vector3 &force, float g = 0.0, int depth = 0);


	NgObject	*_obj;

	NogredexDemoFrameListener *_nogre_env;

	float _thinktimer;
	Vector3 _force;
};

