/*	Ballz.cpp
 *
 *	Robert Rose June 17, 2005
 *
 *	Copyright (c) 2005 Robert Rose
 *
 *	Permission is granted to do whatever you want with this code as long
 *	as the author is given due credit.
 *	
 */


#include "Nogredex.h"
#include <Ogre.h>

#include "Ballz.h"

#include "Obstacle.h"

Ballz::Ballz(RenderWindow *w, Camera *c, SceneManager *sceneMgr)
: NogredexDemoFrameListener(w, c, sceneMgr)
{
	mSceneMgr->setSkyBox(true, "Examples/StormySkyBox");

	_player = NULL;

	_ballmove = false;

	_desiredcamera = Vector3::ZERO;

	_strategy = AgentSearchStrategyRayCombo;

	mDebug = new DebugGraphics(sceneMgr);
    Root::getSingleton().addFrameListener(mDebug);

	showOverlay(true);

	createScene1();

	
}

Ballz::~Ballz()
{
	destroy();
}

void Ballz::destroy()
{
	if(_player)
		delete _player;

	_agents.clear();
	_obstacles.clear();
}

void Ballz::updateStats()
{
	NogredexDemoFrameListener::updateStats();

	if(_player == NULL)
		// nothing to update!
		return;

	String strat = "";
	if(_strategy == AgentSearchStrategyNaiveGrid)
		strat = "Strategy: Naive Grid, 8 Branches";
	if(_strategy == AgentSearchStrategyNaiveGrid4)
		strat = "Strategy: Naive Grid, 4 Branches";
	if(_strategy == AgentSearchStrategyRays)
		strat = "Strategy: Rays";
	if(_strategy == AgentSearchStrategyRayCombo)
		strat = "Strategy: Combination Ray/Grid";

	String keys = "Keys: w/a/s/d/space (move), n (change strategy)";

	OverlayManager::getSingleton().getOverlayElement("Agent/Fitness")->setCaption("A* Pathfinding Strategies");
	OverlayManager::getSingleton().getOverlayElement("Agent/p1")->setCaption(strat);
    OverlayManager::getSingleton().getOverlayElement("Agent/p2")->setCaption(keys);


}

//	I'm just borrowing an overlay from a previous project.. it works
//
void Ballz::showOverlay(bool show)
{
    Overlay* o = (Overlay*)OverlayManager::getSingleton().getByName("Learning/AgentOverlay");
    if (!o)
		Ogre::Exception( Exception::ERR_ITEM_NOT_FOUND, "Could not find overlay Learning/AgentOverlay", "showLearningOverlay" );
    if (show) {
        o->show();
    }
    else {
        o->hide();
    }
}

void Ballz::createScene0()
{	
	mNogredex->init();

	destroy();

	mNogredex->setGravity(Vector3(0, -9.8, 0));
	addPlane("Material/grid", MAT_DEFAULT);

	_player = new CAgent(this);
}


void Ballz::createScene1()
{
	createScene0();

	for(int i = 0; i < 3; i++)
	{
		CAgent *a = new CAgent(this, AgentTypeEnemy);
		a->setPosition(Vector3(rand() % 30, 5.0, rand() % 30));

		_agents.push_back(a);
	}

	
	for(int i = -3; i < 3; i++)
	{
		for(int k = -3; k < 3; k++)
		{
			if(rand() % 2)
				_obstacles.push_back(new CObstacle(this, Vector3(k * 12, 1, i * 12 + rand() % 5), Vector3(4,1,1)));
			else
				_obstacles.push_back(new CObstacle(this, Vector3(k * 12 + rand() % 5, 1, i * 12), Vector3(1,1,4)));
		}
	}
}

void Ballz::createScene2()
{
	createScene0();

	for(int i = 0; i < 3; i++)
	{
		CAgent *a = new CAgent(this, AgentTypeEnemy);
		a->setPosition(Vector3(rand() % 30, 5.0, rand() % 30));

		_agents.push_back(a);
	}

	
	for(int i = -3; i < 3; i++)
	{
		for(int k = -3; k < 3; k++)
		{
			if(rand() % 2)
				_obstacles.push_back(new CObstacle(this, Vector3(k * 10, 1, i * 10 + rand() % 5), Vector3(5,1,1)));
			else
				_obstacles.push_back(new CObstacle(this, Vector3(k * 10 + rand() % 5, 1, i * 10), Vector3(1,1,4)));
		}
	}
}

void Ballz::createScene3()
{
	createScene0();

	CAgent *a = new CAgent(this, AgentTypeEnemy);
	a->setPosition(Vector3(0,2,10));

	_agents.push_back(a);

	
	_obstacles.push_back(new CObstacle(this, Vector3(0,1,5), Vector3(5,1,1)));

}

bool Ballz::frameStarted(const FrameEvent &evt)
{
    if (!NogredexDemoFrameListener::frameStarted(evt))
	{
        return false;
    }

	float delta = evt.timeSinceLastFrame;

	Vector3 cameraDir = mCamera->getDirection();
	
	Vector3 planeDir = cameraDir;
	planeDir.y = 0.0;
	planeDir.normalise();

	Vector3 upDir(0.0, 1.0, 0.0);

	// move desired camera position next to the player
	_desiredcamera += (_player->getPosition() - _desiredcamera) * delta * 2.0;

	Vector3 newcamera = _desiredcamera + cameraDir * -20.0;
	newcamera.y += 5.0;
	
	mCamera->setPosition(newcamera);

	// determine bounding boxes of obstacles
	for(unsigned int i = 0; i < _obstacles.size(); i++)
	{
	}

	// agent movement
	for(unsigned int i = 0; i < _agents.size(); i++)
	{
		vector< pair<Vector3, Vector3> > visForces;

		_agents[i]->think(delta, _strategy, _agents, *_player, _obstacles, visForces, _ballmove);

		for(unsigned int j = 0; j < visForces.size(); j++)
		{
			mDebug->setColour(ColourValue(0.0, 1.0, 0.0));
			mDebug->addLine(visForces[j].first, visForces[j].second);
		}
	}


	// player movement
	if(mInputDevice->isKeyDown(KC_W)) {
		_player->push(planeDir);
	}
	if(mInputDevice->isKeyDown(KC_S)) {
		_player->push(-planeDir);
	}
	if(mInputDevice->isKeyDown(KC_A)) {
		Vector3 left = -planeDir.crossProduct(upDir);
		_player->push(left);
	}
	if(mInputDevice->isKeyDown(KC_D)) {
		Vector3 right = planeDir.crossProduct(upDir);
		_player->push(right);
	}
	if(mInputDevice->isKeyDown(KC_SPACE)) {
		if(_player->getPosition().y < 1.3)
			_player->push(upDir, 20000.0);
	}
	

	if(mInputDevice->isKeyPressed(KC_N)) {
		_strategy = (AgentSearchStrategy) ((_strategy + 1) % 4);
	}
	if(mInputDevice->isKeyPressed(KC_RETURN)) {
		_ballmove = !_ballmove;
	}

	if(mInputDevice->isKeyPressed(KC_1)) {
		createScene1();
	}
	if(mInputDevice->isKeyPressed(KC_2)) {
		createScene2();
	}
	if(mInputDevice->isKeyPressed(KC_3)) {
		createScene3();
	}

	return true;
}